#pragma warning(disable:4244)
#pragma warning(disable:4018)

#include "definitions.h"
#include "textoutput.h"
#include "particle.h"
#include "tetramesh.h"

DclVector getWorld();
void  loadTetraFile(Mesh& m, char *filename);
void  computeSpringSet(Mesh& m);


const int MAX_NUM_PARTICLES = 10000;

//////////// GLOBALS /////////////////
std::vector<Mesh> vectorMeshes;

int frame = 0;

//////////////////////////////////////

void  visualizeParticles();

//------------------------------------------------------------------------------------------------------------------------------------

void  initializeTetraMesh(Mesh& m, DclVector translation)
{
  for (int p = 0; p < m.vectorParticles.size(); p ++)
  {
    Particle& particle = m.vectorParticles[p];
    particle.position += translation;
    particle.force.clear();
    particle.velocity.clear();
    particle.mass = 1.0;
    particle.r = m.r; particle.g = m.g; particle.b = m.b;
  }
  for (int s = 0; s < m.vectorSprings.size(); s ++)
  {
    Spring& spring = m.vectorSprings[s];
    spring.initialLength = (m.vectorParticles[spring.p1].position - m.vectorParticles[spring.p2].position).length();
  }
  for (int t = 0; t < m.vectorTetrahedra.size(); t ++)
  {
    Tetrahedron& tetrahedron = m.vectorTetrahedra[t];
	DclVector p1 = m.vectorParticles[tetrahedron.p1].position;
	DclVector p2 = m.vectorParticles[tetrahedron.p2].position;
	DclVector p3 = m.vectorParticles[tetrahedron.p3].position;
	DclVector p4 = m.vectorParticles[tetrahedron.p4].position;
	DclVector e1 = p2-p1, e2=p3-p1, e3=p4-p1; 
    tetrahedron.initialVolume = e1*(e2%e3)/6.0;
  }
}


//------------------------------------------------------------------------------------------------------------------------------------

// initialize the particle properties
void  initializeSimulation(int argc, char ** argv)
{
  char  *param;
  char  defaultParam[] = "cube.dcstructure";

  if (argc > 1) 
    param = argv[1];
  else
    param = defaultParam;

  vectorMeshes.push_back(Mesh());
  vectorMeshes.push_back(Mesh());
  vectorMeshes.push_back(Mesh());
  vectorMeshes.push_back(Mesh());

  Mesh& m1 = vectorMeshes[0];
  m1.integration = Euler;
  m1.timestep = 0.002;
  m1.world_elasticity = 1;
  m1.spring_constant = 1000.0;
  m1.spring_damping_constant = 5.0;
  m1.volume_constant = 1000.0;
  loadTetraFile(m1, param);
  computeSpringSet(m1);
  initializeTetraMesh(m1, DclVector(-3, 0, 0));

  Mesh& m2 = vectorMeshes[1];
  m2.integration = EulerCromer;
  m2.timestep = 0.002;
  m2.world_elasticity = 1;
  m2.spring_constant = 1000.0;
  m2.spring_damping_constant = 20.0;
  m2.volume_constant = 1000.0;
  loadTetraFile(m2, param);
  computeSpringSet(m2);
  initializeTetraMesh(m2, DclVector(-1, 0, 0));

  Mesh& m3 = vectorMeshes[2];
  m3.integration = Verlet;
  m3.timestep = 0.002;
  m3.world_elasticity = 1;
  m3.spring_constant = 1000.0;
  m3.spring_damping_constant = 20.0;
  m3.volume_constant = 1000.0;
  loadTetraFile(m3, param);
  computeSpringSet(m3);
  initializeTetraMesh(m3, DclVector(1, 0, 0));

  Mesh& m4 = vectorMeshes[3];
  m4.integration = RungeKutta4;
  m4.timestep = 0.002;
  m4.world_elasticity = 1;
  m4.spring_constant = 1000.0;
  m4.spring_damping_constant = 5.0;
  m4.volume_constant = 1000.0;
  loadTetraFile(m4, param);
  computeSpringSet(m4);
  initializeTetraMesh(m4, DclVector(3, 0, 0));
}

//------------------------------------------------------------------------------------------------------------------------------------

// add external forces to the sum of forces acting on the particle
void  computeForces(Mesh& m)
{
	// reset the force vector

    for (int part = 0; part < m.vectorParticles.size(); part ++)
    {
      m.vectorParticles[part].force.clear();
    }

	// gravity

	for (int part = 0; part < m.vectorParticles.size(); part ++)
	{
		Particle& particle = m.vectorParticles[part];
		particle.force += DclVector(0, 1.0, 0) * particle.mass * -9.81; // gravitational force
	}

	// spring forces

	for (int spri = 0; spri < m.vectorSprings.size(); spri ++)
	{
		Spring& spring = m.vectorSprings[spri];
		Particle& particle1 = m.vectorParticles[spring.p1], &particle2 = m.vectorParticles[spring.p2];

		DclVector springForceDirection = particle1.position - particle2.position;
		springForceDirection.normalize();
		DclFloat  springForceMagnitude = ((particle1.position - particle2.position).length() - spring.initialLength) * m.spring_constant;
		particle1.force -= springForceDirection * springForceMagnitude;
		particle2.force += springForceDirection * springForceMagnitude;
	}
	
	// volume force

	for (int tetr = 0; tetr < m.vectorTetrahedra.size(); tetr ++)
	{
		Tetrahedron& tetrahedron = m.vectorTetrahedra[tetr];
		Particle& particle1 = m.vectorParticles[tetrahedron.p1]; 
		Particle& particle2 = m.vectorParticles[tetrahedron.p2]; 
		Particle& particle3 = m.vectorParticles[tetrahedron.p3]; 
		Particle& particle4 = m.vectorParticles[tetrahedron.p4]; 
		DclVector p1 = particle1.position;
		DclVector p2 = particle2.position;
		DclVector p3 = particle3.position;
		DclVector p4 = particle4.position;
		DclVector e1 = p2-p1, e2=p3-p1, e3=p4-p1;
		DclFloat magnitude = m.volume_constant*(e1*(e2%e3)/6.0 - tetrahedron.initialVolume);
		DclVector direction = (e2-e1)%(e3-e1);
		particle1.force += direction*magnitude; 
		direction = e3%e2;
		particle2.force += direction*magnitude; 
		direction = e1%e3;
		particle3.force += direction*magnitude; 
		direction = e2%e1;
		particle4.force += direction*magnitude; 
	}

	// spring damping force using predicted velocities

	for (int spri = 0; spri < m.vectorSprings.size(); spri ++)
	{
		Spring& spring = m.vectorSprings[spri];
		Particle& particle1 = m.vectorParticles[spring.p1], &particle2 = m.vectorParticles[spring.p2];

		DclVector v_rel = particle1.velocity + particle1.force*(m.timestep/particle1.mass) - particle2.velocity - particle2.force*(m.timestep/particle2.mass);
		DclVector springForceDirection = particle1.position - particle2.position ;
		DclVector v_rel_proj = v_rel.projectOn(springForceDirection); // project relative velocity onto direction of spring
		springForceDirection.normalize();
		DclFloat  magnitude = v_rel.length() * m.spring_damping_constant;
		particle1.force -= v_rel_proj * magnitude;
		particle2.force += v_rel_proj * magnitude;
	}

}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one time-integration step
void  integrate(Mesh& m)
{

	for (int part = 0; part < m.vectorParticles.size(); part ++)
	{
		Particle& particle = m.vectorParticles[part];

		switch (m.integration)
		{
		case Euler:
			particle.position = particle.position + particle.velocity * m.timestep;
			particle.velocity = particle.velocity + particle.force / particle.mass * m.timestep;
			break;

		case EulerCromer:
			particle.velocity = particle.velocity + particle.force / particle.mass * m.timestep;
			particle.position = particle.position + particle.velocity * m.timestep;
			break;

		case Verlet:
			{
				DclVector positionOld = particle.position;
				particle.position += (particle.velocity + particle.force / particle.mass * m.timestep) * m.timestep;
				particle.velocity = (particle.position - positionOld) / m.timestep;
				break;
			}

		default:
			break;
		}
	}
}

void  integrationRungeKutta(Mesh& m)
{
	computeForces(m);

	for (int part = 0; part < m.vectorParticles.size(); part ++)
	{
		Particle& particle = m.vectorParticles[part];
		particle.positionOld = particle.position;
		particle.velocityOld = particle.velocity;

		particle.k1 = particle.velocity;
		particle.l1 = particle.force / particle.mass;
		particle.k2 = particle.velocityOld + particle.l1 * m.timestep / 2.0;

		particle.position = particle.positionOld + particle.k1 * m.timestep / 2.0;
		particle.velocity = particle.k2;
	}

	computeForces(m);

	for (int part = 0; part < m.vectorParticles.size(); part ++)
	{
		Particle& particle = m.vectorParticles[part];
		particle.l2 = particle.force / particle.mass;
		particle.k3 = particle.velocityOld + particle.l2 * m.timestep / 2.0; 
		particle.position = particle.positionOld + particle.k2 * m.timestep / 2.0;
		particle.velocity = particle.k3;
	}

	computeForces(m);

	for (int part = 0; part < m.vectorParticles.size(); part ++)
	{
		Particle& particle = m.vectorParticles[part];
		particle.l3 = particle.force / particle.mass;
		particle.k4 = particle.velocityOld + particle.l3 * m.timestep;

		particle.position = particle.positionOld + particle.k3 * m.timestep;
		particle.velocity = particle.k4;
	}

	computeForces(m);

	for (int part = 0; part < m.vectorParticles.size(); part ++)
	{
		Particle& particle = m.vectorParticles[part];
		particle.l4 = particle.force / particle.mass;
		particle.position = particle.positionOld + (particle.k1 + particle.k2 * 2 + particle.k3 * 2 + particle.k4) * m.timestep / 6.0;
		particle.velocity = particle.velocityOld + (particle.l1 + particle.l2 * 2 + particle.l3 * 2 + particle.l4) * m.timestep / 6.0;
	}
}

//------------------------------------------------------------------------------------------------------------------------------------

// collision handling with the world
void  worldCollision(Mesh& m)
{
	DclVector world = getWorld();

	world *= 0.5;

	for (int part = 0; part < m.vectorParticles.size(); part ++)
	{
		Particle& particle=m.vectorParticles[part];

		if (particle.position.x < -world.x || particle.position.x > world.x)
		{
			particle.velocity.x *= -m.world_elasticity;
			particle.position.x = SIGN(particle.position.x) * world.x;
		}

		if (particle.position.y < -world.y || particle.position.y > world.y)
		{
			particle.velocity.y *= -m.world_elasticity;
			particle.position.y = SIGN(particle.position.y) * world.y;
		}

		if (particle.position.z < -world.z || particle.position.z > world.z)
		{
			particle.velocity.z *= -m.world_elasticity;
			particle.position.z = SIGN(particle.position.z) * world.z;
		}
	}
}

//------------------------------------------------------------------------------------------------------------------------------------

// compute and output the total energies in the scene
void  computeEnergies(Mesh& m, int i)
{
  DclFloat  e_kin = 0, e_pot = 0, e_def = 0;
  DclVector world = getWorld();

  for (int part = 0; part < m.vectorParticles.size(); part ++)
  {
    Particle& particle = m.vectorParticles[part];

    e_kin += particle.velocity.length2() * 0.5 * particle.mass;
    e_pot += (particle.position.y + world.z * 0.5) * particle.mass * 9.81;
  }

  for (int s = 0; s < m.vectorSprings.size(); s ++)
  {
    Spring& spring = m.vectorSprings[s];
    Particle& particle1 = m.vectorParticles[spring.p1], &particle2 = m.vectorParticles[spring.p2];
    DclFloat  springForceMagnitude = (particle1.position - particle2.position).length() - spring.initialLength;

	e_def += 0.5 * m.spring_constant * springForceMagnitude * springForceMagnitude;
  }

  Textout::printvar(6 + i, 0, "Ekin %f, Epot %f, Edef %f, Sum %f", e_kin, e_pot, e_def, e_kin + e_pot + e_def);
}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one simulation step
void  simulationStep()
{

	for (int mesh = 0; mesh < vectorMeshes.size(); mesh ++)
	{

		Mesh& m = vectorMeshes[mesh];

		worldCollision(m);

		if (m.integration == RungeKutta4)
		{
			integrationRungeKutta(m);
		}
		else
		{
			computeForces(m);
			integrate(m);
		}

		computeEnergies(m, mesh);

	}
}


